ADJECTIVES = [
    "adorable",
    "adventurous",
    "affectionate",
    "ambitious",
    "amusing",
    "anxious",
    "artistic",
    "attentive",
    "beautiful",
    "bold",
    "brave",
    "bright",
    "brilliant",
    "calm",
    "careful",
    "cautious",
    "charming",
    "cheerful",
    "clever",
    "compassionate",
    "confident",
    "considerate",
    "cooperative",
    "courageous",
    "courteous",
    "creative",
    "curious",
    "daring",
    "delightful",
    "determined",
    "diligent",
    "eager",
    "efficient",
    "empathetic",
    "energetic",
    "enthusiastic",
    "friendly",
    "funny",
    "generous",
    "gentle",
    "graceful",
    "grateful",
    "hardworking",
    "helpful",
    "honest",
    "hopeful",
    "humble",
    "imaginative",
    "independent",
    "innovative",
    "insightful",
    "intelligent",
    "intuitive",
    "inventive",
    "jovial",
    "joyful",
    "kind",
    "knowledgeable",
    "loving",
    "loyal",
    "meticulous",
    "modest",
    "motivated",
    "observant",
    "optimistic",
    "organized",
    "passionate",
    "patient",
    "peaceful",
    "perceptive",
    "persistent",
    "playful",
    "polite",
    "practical",
    "proactive",
    "punctual",
    "quick-witted",
    "reliable",
    "resilient",
    "respectful",
    "romantic",
    "sensitive",
    "sincere",
    "skillful",
    "sociable",
    "spirited",
    "spontaneous",
    "supportive",
    "sympathetic",
    "thoughtful",
    "tidy",
    "trustworthy",
    "understanding",
    "unique",
    "versatile",
    "warm",
    "witty",
    "wise",
    "zany",
    "zealous",
]

NOUNS = [
    "apple",
    "backpack",
    "balloon",
    "banana",
    "beach",
    "bear",
    "bed",
    "bicycle",
    "bird",
    "blanket",
    "boat",
    "book",
    "bottle",
    "bridge",
    "butterfly",
    "camera",
    "candle",
    "car",
    "castle",
    "cat",
    "chair",
    "cheese",
    "city",
    "cloud",
    "computer",
    "couch",
    "cow",
    "cup",
    "deer",
    "dog",
    "dolphin",
    "door",
    "dragon",
    "dream",
    "dress",
    "elephant",
    "factory",
    "fire",
    "fish",
    "flower",
    "forest",
    "fox",
    "frog",
    "garden",
    "giraffe",
    "glasses",
    "guitar",
    "hat",
    "horse",
    "house",
    "island",
    "jacket",
    "jellyfish",
    "kangaroo",
    "key",
    "kite",
    "lamp",
    "leaf",
    "lion",
    "lizard",
    "mountain",
    "mouse",
    "mug",
    "ocean",
    "owl",
    "panda",
    "parrot",
    "pen",
    "pencil",
    "phone",
    "piano",
    "pillow",
    "planet",
    "plate",
    "pocket",
    "rabbit",
    "rainbow",
    "river",
    "robot",
    "rocket",
    "rose",
    "sandwich",
    "scarf",
    "shoe",
    "skateboard",
    "snake",
    "snowflake",
    "sock",
    "spoon",
    "star",
    "sun",
    "sunglasses",
    "table",
    "teapot",
    "teddy",
    "tiger",
    "tree",
    "turtle",
    "umbrella",
    "violin",
    "watch",
    "watermelon",
    "whale",
    "window",
    "zebra",
]
